/*
    Copyright (C) 1995 1996 1997 1998 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
*/
/* Umlaute in Langform wandeln */

#ifndef INCLUDE_ME
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "mausconlib.h"
#include "uostr.h"
#include "entumlauten.h"
#endif

#ifndef INCLUDE_ME
#define FUNC(x) x
static unsigned char tbl[512];
static int init_done;
#define MKUCHAR(x) ((unsigned char) x)

static void 
init_uwan (void)
{
	unsigned int i;

	for (i = 0; i < 256; i++)
	{
		tbl[i] = (unsigned char) i;
	}

	tbl[MKUCHAR ('') * 2] = MKUCHAR ('u');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('U');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('a');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('A');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('o');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('O');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('e');
	tbl[MKUCHAR ('') * 2] = MKUCHAR ('s');
	tbl[MKUCHAR ('') * 2 + 1] = MKUCHAR ('s');

	init_done = 1;
}
size_t 
entumlauten (const char *from, char *to)
{
	const unsigned char *f = (const unsigned char *) from;
	int i;
	unsigned char c;
	const char *old_to=to;

	if (!init_done)
		init_uwan ();

	while (*f)
	{
		i = (*f++);
	   /* Tabellenzugriffe sind teuer, also vermeiden */
		if (i < 128)
			*to++ = i;
		else
		{
			i *= 2;
			*to++ = tbl[i++];
			c = tbl[i++];
			if (c != '\0')
				*to++ = c;
		}
	}
	*to = 0;
	return to-old_to;
}

size_t
entumlauten_fast (const char *from, char *to, size_t l)
{
	const unsigned char *f = (const unsigned char *) from;
	int i;
	unsigned char c;
	const char *old_to=to;

	if (!init_done)
		init_uwan ();

	if (!l)
	{
		*to=0;
		return 0;
	}

#define WORK { \
		i=(*f++); \
		if (i<128) *to++=i; \
		else { i *=2 ; *to++ = tbl[i++]; c=tbl[i++]; if (c) *to++=c; } \
	}

   /*
    * Duffs device - loop unrolling.
    */
	switch (l % 16)
	{
		while (*f)
		{
	case 0:
			WORK
	case 15:
			WORK
	case 14:
			WORK
	case 13:
			WORK
	case 12:
			WORK
	case 11:
			WORK
	case 10:
			WORK
	case 9:
			WORK
	case 8:
			WORK
	case 7:
			WORK
	case 6:
			WORK
	case 5:
			WORK
	case 4:
			WORK
	case 3:
			WORK
	case 2:
			WORK
	case 1:
			WORK
		}
	}
#undef WORK
	*to = 0;
	return to-old_to;
}

size_t
entumlauten2uostr_add_fast(const char *from, uostr_t *to,size_t l)
{
	uostr_needmore(to,l*2);
	l=entumlauten_fast(from,to->data+to->len,l);
	to->len+=l;
	return l;
}

size_t
xentumlauten2uostr_add_fast (const char *from, uostr_t *to,size_t l)
{
	uostr_xneedmore(to,l*2);
	l=entumlauten_fast(from,to->data+to->len,l);
	to->len+=l;
	return l;
}
#endif

char *
FUNC(entumlauten1) (const char *from)
{
	static char buf[258];		/* die meisten Strings sind kurz */
	char *res;
	size_t l;

	if (!init_done)
		init_uwan ();
	l = strlen (from);
	if (l * 2 >= sizeof (buf))
	{
		char *to;

		to = (char *)malloc (l * 2 + 1);
		if (!to)
			return NULL;
		entumlauten_fast (from, to, l);
		res = strdup (to);
		free (to);
	}
	else
	{
		entumlauten_fast (from, buf, l);
		res = strdup (buf);
	}
	return res;
}



size_t
FUNC(entumlauten2uostr_dup_fast) (const char *from, uostr_t *to,size_t l)
{
	uostr_cut(to,0);
	return FUNC(entumlauten2uostr_add_fast) (from, to,l);
}

size_t
FUNC(entumlauten2uostr_dup) (const char *from, uostr_t *to)
{
	size_t l=strlen(from);
	uostr_cut(to,0);
	return FUNC(entumlauten2uostr_add_fast) (from,to,l);
}

size_t
FUNC(entumlauten2uostr_add) (const char *from, uostr_t *to)
{
	size_t l=strlen(from);
	return FUNC(entumlauten2uostr_add_fast) (from,to,l);
}

#ifndef INCLUDE_ME
#  define INCLUDE_ME
#  define malloc XMALLOC
#  define strdup XSTRDUP
#  define uostr_dup_cstr uostr_xdup_cstr
#  undef FUNC
#  define FUNC(y) x##y
#  include "entumlauten.c"
#endif
