/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1995 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   rnews.c: Routinen fr das rnews - Format

   Die Hlfte hier ist von INN geklaut und bis zur Unkenntlichkeit 
   verndert ...
*/

#include "config.h"
#include "mausconsys.h"

#include <stdio.h>
#include <conflib.h>
#include <syslog.h>
#include <assert.h>

#include "mauscon.h"
#include "mausconlib.h"
#include "metamsg.h"
#include "rfc.h"
#include "msg_io.h"

	/* rnews-Zeilenende */
#define CRLF "\n"
#define BATCHLEN (512 * 1024)
static char UNPACK[] = "news-unpack";
#define RNEWS_MAGIC1               '#'
#define RNEWS_MAGIC2               '!'
#define MAX_FORKS                  10

#define STRING(x) #x

static int start_child (int fd, const char *path, char * const argv[]);
static int read_line(char *p, size_t size, int fd);


struct rnews_private
{
	FILE *out;
	int in; /* fd */
	BOOLEAN hot;
	BOOLEAN receiver;
	BOOLEAN really_one;
	BOOLEAN eof;
	/* Die erste Zeile eines Newsbatches oder einer normalen Nachricht.
	 * Kann leer sein, wenn ein Batch entkomprimiert wurde.
	 * (#! rnews ... oder irgendetwas anderes, aber *nie* #! cunbatch)
	 * die ist absichtlich ziemlich gross - moeglicherweise steht
	 * da eine Path:-Zeile drin.
	 */
	char lbuf[16300];
	uostr_t mbuf;
	char *fname;
};

static BOOLEAN rnews_in_open(struct rnews_private *priv,uostr_t *errbuf);
static mio_retcode_t rnews_close(mio_t *chan);
static mio_retcode_t rnews_export(mio_t *chan,metamsg_t *);
static mio_retcode_t rnews_import(mio_t *chan,metamsg_t *);
static mio_retcode_t rnews_open(const char *host, const char *mode,
    mio_t **neu, uostr_t *errbuf);
static mio_retcode_t hot_rnews_open(const char *host, const char *mode,
    mio_t **neu, uostr_t *errbuf);
static int batchcount=0;

mio_base_t mio_rnews=
{
    "rnews",

    rnews_open,
    rnews_import,
    rnews_export,
    rnews_close,

    0
};

mio_base_t mio_hot_rnews=
{
    "hot-rnews",

    hot_rnews_open,
    rnews_import,
    rnews_export,
    rnews_close,

    0
};

static mio_retcode_t 
rnews_close(mio_t *chan)
{
	struct rnews_private *priv=(struct rnews_private *) chan->priv_data;
	if (priv->out)
		fclose(priv->out);
	if (priv->in!=-1)
		close(priv->in);
	free(priv->fname);
	uostr_freedata(&priv->mbuf);
	free(priv);

	chan->priv_data=NULL;
	mio_free(chan);
	return MIO_OK;
}

static mio_retcode_t
rnews_open_sub(const char *fname, mio_t **ret_handle, uostr_t *errbuf)
{
	struct rnews_private *priv;
	mio_t *neu;


	neu=mio_alloc(&mio_rnews);
	priv=malloc(sizeof(struct rnews_private));
	if (priv)
		priv->fname=strdup(fname);

	if (!neu || !priv || !priv->fname)
	{
		uostr_xdup_cstr(errbuf,_("out of memory"));
		uostr_x0(errbuf);
		log(LOG_ERR,_("out of memory"));
		if (priv)
			free(priv->fname);
		free(priv);
		free(neu);
		return MIO_RETRY;
	}
	priv->out=NULL;
	priv->hot=FALSE;
	priv->receiver=FALSE;
	priv->really_one=FALSE;
	priv->eof=FALSE;
	priv->lbuf[0]=0;
	priv->in=-1;
	priv->mbuf.data=NULL;
	priv->mbuf.len=0;
	priv->mbuf.size=0;

	neu->priv_data=priv;
	*ret_handle=neu;
	return MIO_OK;
}

static mio_retcode_t
rnews_open(const char *fname, const char *mode, mio_t **ret_handle, uostr_t *errbuf)
{
	struct rnews_private * priv;
	mio_retcode_t ret;

	if (strcmp(mode,"r") && strcmp(mode,"w"))
	{
		uostr_xdup_cstrmulti(errbuf,"illegal mode '",mode,"' for rnews_open");
		uostr_x0(errbuf);
		log(LOG_ERR,"illegal mode '%s' for rnews_open",mode);
		return MIO_ERROR;
	}

	ret=rnews_open_sub(fname,ret_handle,errbuf);
	if (ret!=MIO_OK)
		return ret;

	priv=(*ret_handle)->priv_data;
	priv->hot=FALSE;

	if (!strcmp(mode,"r"))
	{
#define STARTSIZE (64*1024L)
		uostr_xneed(&priv->mbuf,STARTSIZE);

		if (rnews_in_open(priv, errbuf))
			return MIO_OK;
		log(LOG_ERR,"%s",errbuf->data);
		return MIO_ERROR;
	}
	return MIO_OK;
}

static mio_retcode_t
hot_rnews_open(const char *fname, const char *mode, mio_t **ret_handle, uostr_t *errbuf)
{
	mio_retcode_t ret;

	if (!strcmp(mode,"r"))
		return rnews_open(fname,mode,ret_handle,errbuf);

	ret=rnews_open_sub(fname,ret_handle,errbuf);
	if (ret==MIO_OK)
		((struct rnews_private *)(*ret_handle)->priv_data)->hot=TRUE;
	return ret;
}

static mio_retcode_t 
rnews_export(mio_t *chan, metamsg_t *meta)
{
	struct rnews_private *priv;
	priv=(struct rnews_private *) chan->priv_data;

	if (priv->out)
	{
		if (strlen(meta->body.data)+ftell(priv->out)>BATCHLEN)
		{
			if (priv->hot)
			{
				uostr_t cmd=UOSTR_INIT;
				uostr_xdup_cstrmulti(&cmd,"rnews <",priv->fname,NULL);
				uostr_x0(&cmd);
				if (0==system(cmd.data))
					unlink(priv->fname);
				else
					log(LOG_NOTICE,_("rnews failed on %s"),priv->fname);
				uostr_freedata(&cmd);
			}
			free(priv->fname);
			fclose(priv->out);
			priv->out=NULL;
		}
	}
	if (!priv->out)
	{
		char fname[PATH_MAX];
		char *pos;
		struct stat st;
		long i;
		sprintf (fname, "%s/%s/incoming/newsbatch.", opts.spoolbase, action);
		pos=fname+strlen(fname);

		i=batchcount;
		do
		{
			i++;
			sprintf(pos,"%ld",i);
		} while (stat(fname,&st)==0 && st.st_size!=0);
		batchcount=i;

		priv->out=fopen(fname,"w");
		if (!priv->out)
		{
			int errno1=errno;
			mio_set_errtext(chan,strerror(errno1),NULL);
			log(LOG_ERR,"%s",strerror(errno1));
			return MIO_RETRY;
		}
		priv->fname=XSTRDUP(fname);

		setvbuf (priv->out, malloc(16384), _IOFBF, 16384);
	}	
    /*
     * Das rnews-format benoetigt die Laenge der Nachricht inclusive
     * body. Die haben wir noch gar nicht ...
     * Die rfc_export-Routine soll aber nach Mglichkeit genutzt werden
     * (Vermeidung doppelter Programmierung), also was tun? Eine
     * besonders tolle Mglichkeit ist mir nicht eingefallen, was jetzt
     * kommt, ist definitiv ein Hack ;-)
     */
    /* Viel Platz lassen :-) */
    {
        const char *filler="                    \n";
        long fillerpos;
        long newspos;
        long endpos;
        fputs("#! rnews ",priv->out);
        fillerpos=ftell(priv->out);
        fputs(filler,priv->out);
        newspos=ftell(priv->out);
        rfc_print_message(meta,priv->out,"\n",0);
        endpos=ftell(priv->out);
        if (fseek(priv->out,fillerpos,SEEK_SET))
        {
            /* Tja ... das heit, der Batch ist kaputt. Ende und aus. */
            log (LOG_CRIT,	_("cannot seek on newsbatchf: %s"),
                strerror (errno));
            exit(1);
        }
        fprintf(priv->out,"%ld",endpos-newspos);
        /* Zurueck zum Ende */
        /* Zurueck zum Ende */
        if (fseek(priv->out,endpos,SEEK_SET))
        {
            /* Auch das ist ein unrecoverable error */
            log (LOG_CRIT, _("cannot seek on newsbatchf: %s"),
                strerror (errno));
            exit(1);
        }
    }
    msglog("  %s: newsbatched to batch #%d",meta->message_id,batchcount);
    return 0;
}





/* stolen from INN */
static int 
read_line(char *p, size_t size, int fd)
{
	char    *save;

	/* Fill the buffer, a byte at a time. */
	for (save = p; size > 0; p++, size--) 
	{
		int got=read(fd,p,1);
		if (got==0)
			return -2;
		else if (got != 1) 
		{
			*p = '\0';
			return -1;
		}
		if (*p == '\n') 
		{
			p[1]=0;
			return 0;
		}
	}
	*p = '\0';
	syslog(LOG_ERR, _("line too long: %s"), save);
	return -1;
}

/* stolen from INN */
static int
start_child (int fd, const char *path, char * const argv[])
{
	int     pan[2];
	int     i;
	pid_t   pid;

	/* Create a pipe. */
	if (pipe(pan) < 0) 
	{
		log(LOG_ERR, _("cant pipe for %s: %s"), path,strerror(errno));
		exit(1);
	}

	/* Get a child. */
	for (i = 0; (pid = fork()) < 0; i++) 
	{
		if (i == MAX_FORKS) 
		{
			log(LOG_ERR, _("can't fork %s: %s -- leaving"), path,
				strerror(errno));
			return -1;
		}
		log(LOG_NOTICE, _("can't fork %s -- waiting"), path);
		(void)sleep(60);
	}

	/* Run the child, with redirection. */
	if (pid == 0) 
	{
		(void)close(pan[0]);

		/* Stdin comes from our old input. */
		if (fd != 0) 
		{
			if ((i = dup2(fd, 0)) != 0) 
			{
				log(LOG_ERR, _("can't dup2 %d to 0 got %d %s"), fd, i,
					strerror(errno));
				_exit(1);
			}
			(void)close(fd);
		}

		/* Stdout goes down the pipe. */
		if (pan[1] != 1) 
		{
			if ((i = dup2(pan[1], 1)) != 1) 
			{
				log(LOG_ERR, _("can't dup2 %d to 1 got %d %s"),
					pan[1], i,strerror(errno));
				_exit(1);
			}
			(void)close(pan[1]);
		}
		(void)execv(path, argv);
		log(LOG_ERR, _("can't execv %s %s"), path,strerror(errno));
		_exit(1);
	}

	(void)close(pan[1]);
	(void)close(fd);
	return pan[0];
}

static mio_retcode_t 
rnews_import(mio_t *chan, metamsg_t *meta)
{
	struct rnews_private *priv;
	uostr_t *mbuf;
	long    artsize;

	priv=(struct rnews_private *) chan->priv_data;

	assert(priv->in != -1);
	mbuf=&priv->mbuf;

	if (priv->eof)
		return MIO_EOF;

	if (priv->really_one)
	{
		struct stat st;
		/* das war kein Batch, sondern eine einzelne Nachricht */
		if (fstat(priv->in,&st)==-1)
		{
			int errno1=errno;
			mio_set_errtext(chan,_("cannot fstat on "),
				priv->fname ? priv->fname : "???", 
				": ", strerror(errno1), NULL);
			log(LOG_ERR,_("cannot fstat on %s: %s"),
				priv->fname ? priv->fname : "???", strerror(errno1));
			return MIO_ERROR;
		}
		uostr_xadd_cstr(mbuf,priv->lbuf);

		/* die Gre der Datei kennen wir nicht - das ist ja wahrscheinlich
		 * eine Pipe. */

		if (0!=rfc_read_message_bytes(priv->in, -1, mbuf))
		{
			int errno1=errno;
			mio_set_errtext(chan,_("cannot read from "),
				priv->fname ? priv->fname : "???", 
				": ", strerror(errno1),NULL);
			log(LOG_ERR,_("cannot read from %s: %s"),
				priv->fname ? priv->fname : "???", strerror(errno1));
			return MIO_ERROR;
		}
		priv->eof=TRUE;

		if (rfc_to_meta(mbuf,meta, 0)) /* 0: is_news */
		{
			/* Fehler beim Umwandeln einer Nachricht. Nun gut, allzu
			 * schlimm ist das nicht, weiter mit der nchsten. 
			 * Hufigste Ursache: Kaputter Header, unparsbare Adresse
			 * oder Datum.
			 */
			mio_set_errtext(chan,_("rfc_to_meta failed"),NULL);
			return MIO_RETRY;
		}

		return MIO_OK;
	}
	/* OK. Wir haben einen Batch vorliegen */
	if (priv->lbuf[0]=='\0')
	{
		/* das kann nach cunbatch auftreten - als erste Zeile nach cunbatch */
		/* nchste Zeile lesen */
		switch(read_line(priv->lbuf, (int)(sizeof(priv->lbuf) - 1), priv->in))
		{
		case -2:
			priv->eof=TRUE;
			return MIO_EOF;
		case -1:
			{
				int errno1=errno;
				mio_set_errtext(chan,_("can't read next line: "),
					strerror(errno1),NULL);
				log(LOG_ERR,_("can't read next line: %s"),strerror(errno1));
				return MIO_ERROR; /* naja, fr die falsche Nachricht ... */
			}
		}
	}

	if (strncmp(priv->lbuf, "#! rnews ", 9) != 0) 
	{
		int errno1=errno;
		mio_set_errtext(chan,_("expected '#! rnews ', got "),
			priv->lbuf, ": ", strerror(errno1), NULL);
		log(LOG_ERR,_("expected '#! rnews ', got %s: %s"),
			priv->lbuf, strerror(errno1));
		return MIO_ERROR;
	}
	artsize=strtol(priv->lbuf+9,NULL,10);
	if (artsize <= 0) 
	{
		log(LOG_ERR, _("bad line, bad count: %s"), priv->lbuf);
		mio_set_errtext(chan, _("bad line, bad count: "), priv->lbuf, NULL);
		return MIO_ERROR;
	}
	priv->lbuf[0]='\0';

	mbuf->len=0;

	if (0!=rfc_read_message_bytes(priv->in, artsize, mbuf))
	{
		int errno1=errno;
		mio_set_errtext(chan, _("cannot read from "),
			priv->fname ? priv->fname : "???", 
			": ", strerror(errno1), NULL);
		log(LOG_ERR,_("cannot read from %s: %s"),
			priv->fname ? priv->fname : "???", strerror(errno1));
		return MIO_ERROR;
	}

	if (rfc_to_meta(mbuf,meta,0)) /* 0: is_news */
	{
		/* Fehler beim Umwandeln einer Nachricht. Nun gut, allzu
		 * schlimm ist das nicht, weiter mit der nchsten. 
		 * Hufigste Ursache: Kaputter Header, unparsbare Adresse
		 * oder Datum.
		 */
		return MIO_RETRY;
	}

	return MIO_OK;
}

/* stolen from INN */
static BOOLEAN
rnews_in_open(struct rnews_private *priv,uostr_t *errbuf)
{
	int fd;
	char * cargv[4];

	fd=open(priv->fname,O_RDONLY);
	if (fd==-1)
	{
		uostr_xdup_cstrmulti(errbuf,"can't open batch ",priv->fname,": ",strerror(errno),NULL);
		uostr_x0(errbuf);
		return FALSE;
	}

	/* read the first line */
	if (read_line(priv->lbuf, (int)(sizeof(priv->lbuf)- 1), fd)!=0)
	{
		uostr_xdup_cstrmulti(errbuf,"can't read first line: ",strerror(errno),NULL);
		uostr_x0(errbuf);
		return FALSE;
	}
	
	if (strncmp(priv->lbuf, "#! rnews ", 9) == 0) 
	{
		/* this is a news batch */
		priv->in=fd;
		return TRUE;
	}

	if (strcmp(priv->lbuf, "#! cunbatch\n") != 0) 
	{
		if (priv->lbuf[0] == RNEWS_MAGIC1)
		{
			/* das ist absolut kein Newsbatch und keine einzelne Nachricht */
			priv->lbuf[20]=0;
			uostr_xdup_cstrmulti(errbuf,
				"garbage at start of news batch: starts with '",
					priv->lbuf,"'",NULL);
			uostr_x0(errbuf);
			return FALSE;
		}
		/* Not a batch file. read rest of input as an article. */
		priv->really_one=TRUE;
		priv->in=fd;
		return TRUE;
	}

	/* oh - this is a compressed batch ... */
	cargv[0] = UNPACK;
	cargv[1] = XSTRDUP("-d"); /* get rid of warning */
	cargv[2] = NULL;
	fd = start_child(fd, 
		(opts.compress) ? opts.compress : "gzip", cargv);
	free(cargv[1]);
	if (fd < 0)
	{
		uostr_xdup_cstr(errbuf,"can't start unpacker");
		uostr_x0(errbuf);
		return FALSE;
	}
	priv->in=fd;
	return TRUE;
}

