/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   log.c: Logeintragungen schreiben.
*/

#include "mausconsys.h"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>
#include <syslog.h>

#include "mauscon.h"
#include "mausconlib.h"
#include "uotime.h"

static FILE *msglogfile;

void
open_msglog(void)
{
	uostr_t fname=UOSTR_INIT;
	char date[20];
	uo_sec70_t t;
	uo_datetime_t dt;
	size_t pos1;
	t=uo_now();
	uo_sec702dt(&dt,&t);

	yyyy_mm_dd(date,&dt,'-');
	if (opts.msglog) {
		msglogfile=fopen(opts.msglog,"a");
		if (msglogfile)
			return;
		if (errno==EISDIR) {
			uostr_xdup_cstrmulti(&fname,opts.msglog,"/",date+2,NULL);
			uostr_x0(&fname);
			msglogfile=fopen(fname.data,"a");
			if (msglogfile) {
				uostr_freedata(&fname);
				return;
			}
		}
	}
	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/msglog",NULL);
	pos1=fname.len;
	uostr_xadd_char(&fname,'/');
	uostr_xadd_cstr(&fname,date+2);
	uostr_x0(&fname);
	msglogfile=fopen(fname.data,"a");
	if (!msglogfile && errno==ENOTDIR)
	{
		uostr_cut(&fname,pos1);
		uostr_x0(&fname);
		msglogfile=fopen(fname.data,"a");
	} else if (!msglogfile && errno==ENOENT) {
		uostr_cut(&fname,pos1);
		uostr_x0(&fname);
		mkdir(fname.data,0755);
		uostr_cut(&fname,pos1);
		uostr_xadd_char(&fname,'/');
		uostr_xadd_cstr(&fname,date+2);
		uostr_x0(&fname);
		msglogfile=fopen(fname.data,"a");
	}
	if (!msglogfile)
	{
		log(LOG_ERR,_("cannot open msglog: %s"), strerror(errno));
		exit(1);
	}
	uostr_freedata(&fname);
}

void 
close_msglog(void)
{
	if (msglogfile)
	{
		fclose(msglogfile);
		msglogfile=NULL;
	}
}

void
msglog(const char *format,...)
{
	va_list ap;
	uo_sec70_t t;
	FILE *out;

	static char str[32];
	static uo_sec70_t last_time=0;

	if (!msglogfile)
		out = stderr;
	else
		out = msglogfile;
	va_start (ap, format);

	t=uo_now();
	if (t!=last_time) {
		uo_datetime_t dt;
		uo_sec702dt(&dt,&t);
		yyyy_mm_dd_hh_mm_ss(str,&dt,'-');
		last_time=t;
	}
	fputs(str,out);
	vfprintf (out, format, ap);
	fputc('\n',out);
	fflush (out);
	va_end (ap);
}

void
log (int priority, const char *format,...)
{
	va_list ap;

#ifdef HAVE_VSYSLOG
	va_start (ap, format);
	vsyslog (priority, format, ap);
	va_end (ap);
#else
	char *s;
	va_start (ap, format);
	if (vasprintf(&s,format,ap)<0) {
		va_end (ap);
		syslog(LOG_ERR,"out of memory, syslog message lost");
	} else {
		va_end (ap);
		syslog(priority,"%s",s);
	}
#endif

	/* Kopie der Nachricht nach stderr. Eventuell mal konfigurierbar machen */
	va_start (ap, format);
	vfprintf (stderr, format, ap);
	fputc('\n',stderr);
	va_end (ap);
}
