/* mauscon.h: prototypes and definitions */
/*
    Copyright (C) 1994 1995 1996 1997 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#if !defined(LIB_H) || defined(INIT)
#ifndef INIT
#define LIB_H
extern const char *gate_version;

#include "config.h"

extern const char *program_name;

	/*
	 * Zeilenende fuer MausTausch, SMTP, NNTP
	 */
#define LINEEND "\r\n"
	/*
	 * Erlaubte Zeilenendezeichen. *Nach* diesen Zeichen kommen nur
	 * noch andere Zeilenendezeichen.
	 */
typedef struct mc_list_struct
{
	char *data;
	char *ext;
	struct mc_list_struct *next;
} mc_list_t;

typedef enum 
{ 
	MAUS2RFC=1, 
	RFC2MAUS=2, 
	MROUTE=4,
	GATECONFTEST=8,
	CALL=16,
	GETMAUSVAL=32,
	MCOM=64,
	MHIST=128
} PROGRAM;


/* Defaultbezeichner */
#define CNFNAME "mauscon.cnf"

/* allgemeine Libraryfunktionen */
char *strtrim(char *s);
int strhalfcasecmp(const char *uplow, const char *low);
int strnhalfcasecmp(const char *uplow, const char *low, size_t n);

int do_conf (const char *fname, const char *aktion, 
	char **direct_vars, int setvarcount);
const char *get_conf_value(const char *which);
void dump_conf(FILE *f);
time_t get_date(char *p, time_t now, long zone);

typedef struct cancel_struct
{
	char *id;
	size_t local_part_len; /* Optimierung der Suchroutine */
	char *from;
	struct cancel_struct *next;
	int done;
} cancel_t;

/* canceln von Nachrichten */
int cancel_init(int keep_locked);
cancel_t *cancel_find_in_list(const char *id);
cancel_t *cancel_anker(void);
void cancel_end(int do_delete,int rewrite);
void cancel_add(const char *id, const char *from);

/* Libraryfunktionen fuer Logfile */
void open_msglog(void);
void close_msglog(void);
void log(int priority, const char *format, ...)
     MC_ATTRIB_PRINTF(2,3);
void msglog(const char *format, ...)
     MC_ATTRIB_PRINTF(1,2);


/* Kommandozeilenauswertung */
extern const char *action;
enum work_only_enum { ONLY_ALL, ONLY_MAIL, ONLY_NEWS, ONLY_SPECIAL };
extern int work_only;
extern int dont_use_nntp;
extern int dont_use_smtp;

typedef enum { USER, GATE } WORKMODE;
typedef enum { LOGMODE_BOTH, LOGMODE_SYSLOG, LOGMODE_FILE } LOGMODE;
typedef enum 
{ 
	BOXTYP_NONE,
	BOXTYP_QUARK, 
	BOXTYP_MAUS 
} BOXTYP;

typedef enum 
{ 
	REKILL_FULL,  /* kill all re: */
	REKILL_SMART, /* kill re: if line too long */
	REKILL_NONE   /* don't kill re: */
} REKILLMODE;

/* Message-Trace-Flags. A better place for them is
 * metamsg.h, but then i can't set the defaults in the
 * options structure here
*/
    /* XR- set due to control msg header
     *  (includes Also-Control, Expires, Supersedes)
     * Should not be set together which MMT_CONTROL.
     */
#define MMT_CONTROL      1
    /* distribution lines: anything strange found in distribution
     * lines, any modifications done to it.
     */
#define MMT_DIST         4
    /* some error occured during conversion, which was not expected
     * because the other side should be able to handle this
     */
#define MMT_ERROR        8
    /* Noexport flag set. Set this whenever noexport is set
     * rarely set alone, only if XR- found.
     * does not imply XR-, some errors don't really matter
     */
#define MMT_NX          16
    /* anything do to ids, strange things found there */
#define MMT_ID          32
    /* just information */
#define MMT_INFO        64
    /* path lines: modification done to it */
#define MMT_PATH       128
    /* MIME trouble */
#define MMT_MIME       128
    /* conversion loss: something which is totally OK in one format
     * can't be converted to the other.
     */
#define MMT_LOSS       256

#define HONOR_CANCEL_PASS 1
#define HONOR_CANCEL_SELF 2

#ifdef __GNUC__
#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 6)
#define MC_ATTRIB_UNUSED __attribute__((__unused__))
#endif
#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 5)
# define MC_ATTRIB_PRINTF(formatnr,firstargnr)  \
    __attribute__((__format__ (printf,formatnr,firstargnr)))
#endif
#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4)
#  define MC_ATTRIB_CONST  __attribute__((__const__))
#endif
    /* gcc.info sagt, noreturn wre ab 2.5 verfgbar. HPUX-gcc 2.5.8
     * kann es noch nicht - what's this?
     */
#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 5)
# define MC_ATTRIB_NORET  __attribute__((__noreturn__))
#endif
#endif /* __GNUC__ */
#ifndef MC_ATTRIB_UNUSED
#define MC_ATTRIB_UNUSED
#endif
#ifndef MC_ATTRIB_CONST
#define MC_ATTRIB_CONST
#endif
#ifndef MC_ATTRIB_NORET
#define MC_ATTRIB_NORET
#endif
#ifndef MC_ATTRIB_PRINTF
#define MC_ATTRIB_PRINTF(x,y)
#endif

BOOLEAN is_allowed (const char *sender);



/***************************************************************************
 * Variablen der Konfigurationsdatei
 * genutzt fuer Deklaration und Initialisierung, via INIT-Makro 
 ***************************************************************************/
#define INIT(x,y)	x;
#define IS_DECLARE
extern struct overalloptions {
#else
struct overalloptions opts= {
#endif
	INIT(WORKMODE workmode,                USER)

	INIT(const char *spoolbase,         "/var/spool/mauscon")
#ifdef LOG_NEWS
	INIT(int sl_facility,               LOG_NEWS)
#else
	INIT(int sl_facility,               LOG_DAEMON)
#endif
	INIT(int sl_priority,               LOG_NOTICE)
	INIT(int smart_LOG,                 1)

	INIT(const char *gatemaster,        "postmaster")
	INIT(const char *mysitedomain,      NULL)
	INIT(const char *mailuserdomain,       NULL)
	INIT(mc_list_t *allowed,            NULL)
	INIT(mc_list_t *empty_envelope_for,   NULL)
	INIT(const char *v_for_empty_envelope,   "mailer-daemon")

	/* INIT(BOXTYP boxtyp,                 BOXTYP_MAUS) */
	INIT(const char *boxtypname,            "MAUS")
	INIT(const char *pollsite,             NULL)
	INIT(const char *pollsitedomain,       NULL)
	INIT(const char *mausdomain,        ".maus.de")
	INIT(short tabsize,                 0)
	INIT(long split_after,              15500)
	INIT(short subjectlen,              250)
	INIT(int autoinfosubscribe,         1)
	INIT(int rekillmode,                REKILL_FULL)
	INIT(int send_crosspostings,        0)
	INIT(int send_sender,               1)
	INIT(int send_followup_to,          1)
	INIT(int send_reply_to,             1)
	INIT(int send_path,                 0)
	INIT(int generate_empty_mail_body,  1)
	INIT(int cancel_original,           1)
	INIT(int mail_status_change,           1)
	INIT(char autostatus,              '\0')
	INIT(int strip_domain,              1)
	INIT(int strip_bangpath,            0)
	INIT(int t_online_id_hack,          0)


	INIT(int direct_smtp,                  0)
	INIT(const char *mailhost,             "localhost")
	INIT(const char *reroute_host,      NULL)
	INIT(const char *newshost,             "localhost")
	INIT(const char *path_entry,        "MausNet")
	INIT(const char *gprefix,           "maus")
	INIT(const char *usenet_dist_lokal,    "lokal")
	INIT(const char *usenet_dist_mausnet,  "mausnet")
	INIT(int warn_news_routing,         0)
	INIT(int unlink_unallowed,          1)
	INIT(short max_references,          4)
	INIT(int enable_dsn_bounces,        0)

	INIT(int foreign_export_kill, 0)
	INIT(int allow_transit, 0)
	INIT(const char *foreign_export_path, NULL)
	INIT(const char *foreign_export_distribution, NULL)

	INIT(int mime_coded_headers,        0)
	INIT(int mime_coded_body,           0)
	INIT(int mime_headerlines,          1)

	INIT(const char *rmail_route,       NULL)
	INIT(const char *rsmtp_route,       NULL)
	INIT(const char *rnews_route,       NULL)

	INIT(const char *compress,              "gzip")
	INIT(short lock_tries,              10)
	INIT(short lock_sleep,              5)
	INIT(const char *diffcmd,           "diff -u --report-identical-files")
	INIT(int keep_sent,                 0)

	INIT(unsigned long debugmask_maustausch,  MMT_ERROR|MMT_NX|MMT_ID|MMT_MIME|MMT_LOSS)
	INIT(unsigned long debugmask_rfc,   MMT_ERROR|MMT_NX|MMT_ID|MMT_MIME|MMT_LOSS)

	INIT(const char *quark_user_dir,               NULL)
	INIT(int transparent_ids,            0)
	INIT(int honor_cancels,              HONOR_CANCEL_SELF|HONOR_CANCEL_PASS)
	INIT(int cancel_bx_with_v,           1)
	INIT(int send_short_ids,             0)
	INIT(int send_b_lines,				 0)

	INIT(const char *smtpd_mroute,               NULL)
	INIT(const char *msglog,               NULL)
/* fuer revo */
	INIT(const char *boxkuerzel,           NULL)
	INIT(const char *defaultdomain,        NULL)
#ifdef IS_DECLARE
#undef INIT
#undef IS_DECLARE
	char *dummy;
} opts;
#else
	NULL
};
#endif



#endif
