/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mimesup.h"
#include "mausconlib.h"


static unsigned char *b64table;
static void b64init(void)
{
	int i;
	b64table=(unsigned char *) XMALLOC(256*sizeof(char));
	memset(b64table,255,256);
	for (i='A'; i<='Z'; i++)  b64table[i]=i-'A';
	for (i='a'; i<='z'; i++)  b64table[i]=i-'a'+26;
	for (i='0'; i<='9'; i++)  b64table[i]=i-'0'+52;
	b64table['+']=62;
	b64table['/']=63;
}

/* Is it an supported charset? */
static char *
mime_charset_supported(const char *charset)
{
	char cs[256];
	char *p;
	size_t l;
	if (!charset)	
		return NULL;

	/* 
	 * strtrim(charset) -> handle common problems 
	 */
	l=strlen(charset);
	if (l>=sizeof(cs)-1)
		return NULL;
	memcpy(cs,charset,l+1);
	p=cs+l-1;
	while (isspace(*p) && p>=cs)
		p--;
	p[1]='\0';
	p=cs;
	while(isspace(*p))
		p++;
	if (0==strcasecmp(p,"iso-8859-1")
		|| 0==strcasecmp(p,"us-ascii")
		|| 0==strcasecmp(p,"ascii")
		|| 0==strcasecmp(p,"latin1"))
		return p;
	else
		return NULL;
}

void 
mime_decode_headerline(char *str)
{
	char *start;
	start=str;
	while (*start && (start[0]!='=' || start[1]!='?'))
		start++;
	if (!*start)
		return;
	/* while there may be another encoded word */
	while (start)
	{
		char *charset;
		char *encoding;
		char *text;
		char *end;
		char *got_charset;
		int got_encoding;
		char encodechar;
		/* 
		 * =?charset?encoding?enc.text=?
		 * ^- start
		 *   ^- charset
		 *           ^- encoding
		 *                    ^- test
		 *                            ^- end
		 */
		end=strstr(start,"?=");
		if (!end)
			break; /* can't translate, don't change (further) */
		charset=start+2;
		encoding=strchr(charset+1,'?');
		if (!encoding)
			break;
		encoding++; /* now on Q or B */
		text=strchr(encoding+1,'?');
		if (!text)
			break;

		/* try to find the charset */
		encoding[-1]=0;
		got_charset=mime_charset_supported(charset);
		encoding[-1]='?';

		/* known encoding? */
		encodechar=tolower(*encoding);
		got_encoding = ((encodechar=='q'||encodechar=='b') && encoding[1]=='?');
		
		/* if supported charset && supported encoding */
		if (got_charset && got_encoding && encodechar=='q') {
			/* known charset, quoted printable ->translate */
			/* we do this in place to avoid the use of tmp storage */
			char *p,*q;
			size_t restlen;

			p=text+1;
			for (p=text+1,q=start;*p && p<end;) {
				int c=*p;
				switch(c)
				{
				case '=':
					{
						char buf[3];
						buf[0]=p[1];
						buf[1]=p[2];
						buf[2]='\0';
						c=(int)strtol(buf,NULL,16);
						if (c<=0 || (c<' ' && c!='\t')) {
							memcpy(q,p,3);
							q+=3;
						} else
							*q++=c;
						p+=3;
					}
					break;
				case '_':
					*q++=' ';
					p++;
					break;
				default:
					*q++=c;
					p++;
					break;
				} /* case */
			} /* while */
			/* any 'linear-white-space' that separates a pair of
			   adjacent 'encoded-word's is ignored.
			 */
			p=end+2;
			while (isspace(*p))
				p++;
			if (*p=='=' && p[1]=='?') {
				end=p;
			} else {
				end+=2;
			}
			restlen=strlen(end);
			memmove(q,end,restlen+1);
			start=q;
		}
		else if (got_charset && got_encoding && encodechar=='b')
		{
			/* base64 */
			unsigned char *pr;
			char *pw;
			size_t len;
			long l=0; /* keep gcc happy */
			int r;
			int val;
			char *p;
			if (!b64table)
				b64init();
			pr=(unsigned char *)text+1;
			pw=start;
			r=0;
			while (*pr!='?' && *pr)
			{
				if (*pr=='=')
				{	
					pr++;
					continue;
				}
				val=b64table[*pr++];
				if (val==255)
					continue; /* illegal char, ignore */
				switch(++r)
				{
				case 1:
					l=val*262144;
					continue;
				case 2:
					l+=val*4096;
					continue;
				case 3:
					l+=val*64;
					continue;
				case 4:
					l+=val;
					r=0;
					/* Rcksichtname auf Spezialzeichen ... */
#define PUTONE(x) do {if ((x)<' ' && (x)!='\t') *pw++='?'; else *pw++=x; } while(0)
					PUTONE(l/65536);
					l%=65536;
					PUTONE(l/256);
					PUTONE(l%256);
					l=0;
				}
			}
			/* decode leftovers, but careful */
			if (l/65536)
			{
				PUTONE(l/65536);
				l%=65536;
			}
			if (l>256)
				PUTONE(l/256);
			if (l%256)
				PUTONE(l%256);
			if (*pr=='?')
				pr++;
			if (*pr=='=')
				pr++;

			p=(char *)pr;
			while (isspace(*p))
				p++;
			if (*p=='=' && p[1]=='?') {
				end=p;
			} else {
				end=(char *)pr;
			}
			len=strlen(end);
			memmove(pw,end,len+1); /* +1 wg. \0 */
			start=pw;
		}
		else
			start++;
		/* 
		 * else: cannot translate. possibly another program has any idea,
		 * so leave it as it was. 
		 */
		start=strstr(start,"=?");

		/* hier wre eine Zeichensatzwandlung einzubauen */
	}
}

int
mime_content_text(const char *s)
{
	/* Content-Type: type/subtype [;pararameter]* 
	 * we can deal with "text/plain", and "charset=..." 
	 */

	if (*s == '\n')
		return 0; /* well, assume text/plain and us-ascii */
	/* skip whitespace */
	while (isspace(*s))
		s++;
	/* s now points to "type/subtype" */
	if (!strncasecmp(s,"text/plain",10))
		s+=10;
	else if (!strncasecmp(s,"text",4)) /* common mistake */
		s+=4;
	else
		return 0; /* cannot handle it */
	while (isspace(*s))
		s++;
	if (!*s)
		return 1;
	if (*s++!=';') /* Hey? What are you? */
		return 0;
	return 1;
#if 0
	do
	{
		if (!strncasecmp(s,"charset=",8))
		{
			/* found charset */
			char *t;
			char c;
			t=s+8;
			while (!isspace(*t) && *t!=';')
				t++;
			c=*t;
			*t='\0';
			if (mime_charset_supported(t))
				return 1;
			*t=c;
		}
		s=index(s,';');
		if (!s)
			return 1; /* us-ascii */
		s++;
	}
	while (s && *s);
#endif
}
                                                               
enum mime_encoding
mime_get_encoding(const char *s)
{
	enum mime_encoding enc=unknown_encoding;
	while (isspace(*s))
		s++;
	if (strncasecmp(s, "7bit", 4) == 0)
	{
		enc=seven_bit;
		s += 4;
	}
	else if (strncasecmp(s, "8bit", 4) == 0)
	{
		enc=eight_bit;
		s += 4;
	}
	else if (strncasecmp(s, "binary", 6) == 0)
	{
		enc=binary;
		s += 6;
	}
	else if (strncasecmp(s, "quoted-printable", 16) == 0)
	{
		s += 16;
		enc=quoted_printable;
	}
	else if (strncasecmp(s, "base64", 6) == 0)
	{
		s += 6;
		enc=base_64;
	}
	if (*s == '\0' || isspace(*s) || *s == ';' || *s == '(')
		return enc;
	else
		return unknown_encoding;
}

#define DECIMAL(c) \
	( isdigit(c) ? (c-'0') : \
	(c>='A' && c<='F') ? (c-'A'+10) : \
	(c>='a' && c<='f') ? (c-'a'+10) : 0 )

    /* decode a qp buffer */
void 
mime_decode_qp (char *buf)
{
	char *rpos;
	char *wpos;
	unsigned int c;
	rpos=wpos=buf;
	while (1)
	{
		switch(*rpos)
		{
		case '=':
			/* soft linebreaks? */
			if (!strcmp(rpos+1,"\r\n"))
				rpos+=3;
			else if (rpos[1]=='\n')
				rpos+=2;
			else
			{
				c=(unsigned int) DECIMAL(rpos[1])*16+DECIMAL(rpos[2]);
				if (c==0)
					*wpos++=' ';
				else
					*wpos++=c;
				rpos+=3;
			}
			break;
		case '\0':
			*wpos='\0';
			return;
		case '\n':
			/* Therefore, when decoding a Quoted-Printable
			 * body, any trailing white space on a line must be
			 * deleted, as it will necessarily have been added
			 * by intermediate transport agents.
			 */
			 {
			 	char *p=rpos-1;
			 	if (*p=='\r')
			 		p--;
			 	while (isspace(*p) && *p!='\n')
			 	{
			 		p--;
			 		wpos--;
			 	}
			 }
			*wpos++=*rpos++;
			break;
		default:
			*wpos++=*rpos++;
			break;
		}
	}
}

    /* decode a b64 buffer */
void
mime_decode_base64 (char *buf)
{
	unsigned char *pr;
	char *pw;
	long l=0; /* keep gcc happy */
	int r;
	int val;

	if (!b64table)
		b64init();
	pr=(unsigned char *)buf;
	pw=buf;
	r=0;
	while (*pr)
	{
		if (*pr=='=')
		{	
			break; 
			pr++;
			continue;
		}
		val=b64table[*pr++];
		if (val==255)
			continue; /* illegal char, ignore */
		switch(++r)
		{
		case 1:
			l=val*262144;
			continue;
		case 2:
			l+=val*4096;
			continue;
		case 3:
			l+=val*64;
			continue;
		case 4:
			l+=val;
			r=0;
			*pw++=l/65536;
			l%=65536;
			*pw++=l/256;
			*pw++=l%256;
			l=0;
		}
	}
	/* decode leftovers, but careful */
	if (l/65536)
	{
		*pw++=l/65536;
		l%=65536;
	}
	if (l>256)
		*pw++=l/256;
	if (l%256)
		*pw++=l%256;
	*pw='\0';
}

void
mime_decode_buffer(char *buffer,enum mime_encoding enc)
{
	if (enc==base_64)
		mime_decode_base64(buffer);
	else if (enc==quoted_printable)
		mime_decode_qp(buffer);
}


