/* dummy MTA */

#include <unistd.h>
#include <errno.h>
#include <stdlib.h> /* getenv */
#include <string.h>
#include <sys/wait.h>
#include <fcntl.h>
#include "uostr.h"
#include "uoio.h"
#include "str_ulong.h"
#include "mtalib.h"

static mta_t * mta_dummy_create P__((void));
static mta_ret_t mta_dummy_write P__((mta_t *mta,const char *s, size_t len));
static mta_ret_t mta_dummy_close P__ ((mta_t *mta));
static mta_ret_t mta_dummy_abort P__ ((mta_t *mta));
static mta_ret_t mta_dummy_open P__ ((mta_t **,const char *,const char *, const char *));

struct priv {
	int fd;
	uoio_t io;
	int goterrs;
};
mta_t mta_dummy={
	mta_dummy_open,
	mta_dummy_write,
	mta_dummy_abort,
	mta_dummy_close,
	0
};

static mta_t *
mta_dummy_create(void)
{
	mta_t *m;
	struct priv *p=malloc(sizeof(struct priv));
	if (!p) return 0;
	m=(mta_t *) malloc(sizeof(mta_t));
	if (!m) {free(p); return 0;}
	memcpy(m,&mta_dummy,sizeof(mta_t));
	memset(p,0,sizeof(*p));
	m->priv=p;
	p->fd=-1;
	return m;
}

static mta_ret_t
mta_dummy_open(mta_t **m, const char *path,const char *from, const char *to)
{
	int fd;
	uoio_t *io;
	struct priv *p;
	mta_t *mta=mta_dummy_create();
	if (!mta) { 
		mta_errtxt="out of memory"; 
		mta_hcmssc="(#4.3.0)"; 
		return MTA_DEFER;
	}
	if (!path)
		fd=2;
	else if (0==strcmp(path,"-"))
		fd=1;
	else 
		fd=open(path,O_WRONLY|O_CREAT|O_APPEND,0600);
	if (fd==-1) {
		mta_errtxt=strerror(errno);
		mta_hcmssc="(#4.3.0)"; 
		free(mta->priv);
		free(mta);
		return MTA_DEFER;
	}
	p=mta->priv;
	p->fd=fd;
	io=&p->io;
	uoio_assign_w(io,fd,write,0);
	uoio_write_cstr(io,"--------- Start Metainformation ---------\n");
	if (from) {
		uoio_write_cstr(io,"from: ");
		uoio_write_cstr(io,from);
		uoio_write_cstr(io,"\n");
	} 
	if (to) {
		uoio_write_cstr(io,"to: ");
		uoio_write_cstr(io,to);
		uoio_write_cstr(io,"\n");
	} 
	uoio_write_cstr(io,"--------- End Metainformation, Start Message ---------\n");
	if (-1==uoio_flush(io)) {
		mta_errtxt=strerror(errno);
		mta_hcmssc="(#4.3.0)"; 
		uoio_destroy(io);
		free(mta->priv);
		free(mta);
		return MTA_DEFER;
	}
	*m=mta;
	return MTA_SUCCESS;
}

static mta_ret_t
mta_dummy_write(mta_t *mta,const char *s, size_t len)
{
	uoio_t *io=&(((struct priv *)mta->priv)->io);
	
	if (-1==uoio_write_mem(io,s,len))
		((struct priv *)mta->priv)->goterrs++;
	if (((struct priv *)mta->priv)->goterrs) {
		mta_errtxt=strerror(errno); mta_hcmssc="(#4.3.0)";
		return MTA_DEFER;
	}
	return MTA_SUCCESS;
}

static mta_ret_t
mta_dummy_abort(mta_t *mta)
{
	struct priv *priv=(struct priv *)mta->priv;
	close(priv->fd);
	uoio_destroy(&priv->io);
	free(priv);
	free(mta);
	return MTA_SUCCESS;
}

static mta_ret_t
mta_dummy_close(mta_t *mta)
{
	int goterrs;
	struct priv *priv=mta->priv;
	static char pidbuf[20]; 

	if (priv->goterrs) { mta_errtxt="I/O error"; mta_hcmssc="(#4.3.0)"; }
	uoio_write_cstr(&priv->io,"--------- End Message ---------\n");
	if (-1==uoio_flush(&priv->io)) {priv->goterrs++;mta_errtxt=strerror(errno);mta_hcmssc="(#4.3.0)"; }
	close(priv->io.fd);
	uoio_destroy(&priv->io);
	goterrs=priv->goterrs;

	free(priv);
	free(mta);

	str_ulong(pidbuf,getpid());
	mta_errtxt=pidbuf; /* magic: on MTA_SUCCESS the ..._close() puts an id into the errtxt */
	return MTA_SUCCESS;
}
