#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/file.h>
#include <fcntl.h>
#include <syslog.h>
#include "timselsysdep.h"
#include "uostr.h"
#include "uoio.h"
#include "uotime.h"
#include "mtalib.h"
#include "uogetopt.h"
#include "uo_entumlaut.h"
#include "str_ulong.h"
#include "mausconlib.h"
#include "mauscon.h"
#include "setvars.h"
#define ITZ
#ifdef ITZ
#include "itzlib.h"
#endif
#include "uoconfig.h"

char *myhostname=0;
uoconfstr_t c_realname;
const char *opt_realname;
#ifdef ITZ
uoconfstr_t c_itz;
const char *opt_itz;
#endif
const char *o_usercharset;
uoconfstr_t c_usercharset;

const char *username;

uoconfstr_t c_boxkuerzel;
const char *o_boxkuerzel;
uoconfstr_t c_defaultdomain;
const char *o_defaultdomain;

const char *cnffname=SYSCONFDIR "/" CNFNAME;
const char *rcfname=".tauschrc";
const char *inputfilename;
static const char *opt_LOG;
static int opt_ITG;
static int opt_dry_run; /* XXX */
static int opt_statistics;
static int opt_verbose;

static uogetopt_t myopts[]={
	{'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
	"Channel to use (old name)","CHANNEL"},
	{'C',"channel",UOGO_STRING,&action,0,
	"Channel to use","CHANNEL"},
	{'c',"cnffile",UOGO_STRING,&cnffname,0,
	"Configuration file to use","FILE"},
	{'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&cnffname,0,
	"Configuration file to use","FILE"},
	{'i',"input",UOGO_STRING,&inputfilename,0,
	"Infile file name","FILE"},
	{'I',"ITZ",UOGO_STRING,&opt_itz,0,
	"Use FILE as ITZ\n"
	"Default is to use the ITZ given in $MT_ITZ\n" ,"FILE"},
	{'L',"LOG",UOGO_STRING,&opt_LOG,0,
	"Create LOG" ,"FILE"},
	{'n', "dry-run", UOGO_FLAG, &opt_dry_run, 1,
	"Do change any files, do not send any message\n"
	"external commands will not be run.",0},
	{'O',"OUT",UOGO_STRING,&opt_ITG,0,
	"Save OUT blocks to that file" ,"FILE"},
	{'r',"rcfile",UOGO_STRING,&rcfname,0,
	".tauschrc file\n"
	"This file contains a list of server names,\n"
	"newsgroup names and article number high-water-marks\n"
	"in that groups. Keep off unless you really know what\n"
	"you are doing (in that case you won't touch that file",
	"FILE"},
	{'R',"realname",UOGO_STRING,&opt_realname,0,
	"Set your realname" ,"NAME"},
	{'s',"set",UOGO_CALLBACK,&setvarfn,0,
	"override variables in the configuration file","VAR=VALUE"},
	{'S', "statictics", UOGO_FLAG, &opt_statistics, 1,
	"Collect statistics.\n"
	"Such as: bytes + messages transferred.\n",0},
	{'v', "verbose", UOGO_FLAG, &opt_verbose, 1,
	"Be verbose\n",0},

	{ 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
	"Copyright (C) 1999 Uwe Ohse\n"
	PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
	"to the extent permitted by law.\n"
	"You may redistribute copies of " PACKAGE "\n"
	"under the terms of the GNU General Public License.\n"
	"For more information about these matters, see the file named COPYING.\n"
	"You may redistribute copies of " PACKAGE "\n",0},

	{0,0,0,0,0,0,0} /* --help and --version */
};

static uostr_t headers[128];
static int in_header=0;
static enum messagetype {MT_JUNK,MT_NEWS,MT_MAIL,MT_STATUS,MT_CMD} messagetype=MT_JUNK;
static uoio_t o;
static uoio_t io_log;
static uoio_t input;
static uoio_t e;
static int do_charset_conversion=0;
uo_datetime_t date;
mta_t *mta;

const char *argv0;

static void bailout(int err_no,const char *s1,const char *s2, const char *s3, const char *s4)
{
	write(2,argv0,strlen(argv0)); write(2,": ",2);
	if (s1) write(2,s1,strlen(s1));
	if (s2) write(2,s2,strlen(s2));
	if (s3) write(2,s3,strlen(s3));
	if (s4) write(2,s4,strlen(s4));
	if (err_no) {s1=strerror(errno); write(2,": ",2);write(2,s1,strlen(s1));}
	write(2,"\n",1);
	_exit(1);
}

static 
int hdr_chr(uostr_t *s,char d)
{
	char *c;
	if (s->data && (c=memchr(s->data,d,s->len))) {
		return (c-s->data);
	}
	return -1;
}

static void
write_log(const char *s)
{
	uoio_write_mem(&io_log,":",1); uoio_write_cstr(&io_log,s); uoio_write_mem(&io_log,"\r\n",2);
	uoio_write_cstr(&o,s); uoio_write_mem(&o,"\n",1);
	uoio_flush(&o);
}
static void
write_log2(const char *s, const char *t)
{
	uoio_write_mem(&io_log,":",1); uoio_write_cstr(&io_log,s); uoio_write_cstr(&io_log,t); uoio_write_mem(&io_log,"\r\n",2);
	uoio_write_cstr(&o,s); uoio_write_cstr(&o,t); uoio_write_mem(&o,"\n",1);
	uoio_flush(&o);
}

static int write_transport(const char *s)
{ if (MTA_SUCCESS!=mta->mta_write(mta,s,strlen(s))) return -1; return 0; }
static int write_transport2(const char *s, const char *t)
{ if (MTA_SUCCESS!=mta->mta_write(mta,s,strlen(s))) return -1; 
  if (MTA_SUCCESS!=mta->mta_write(mta,t,strlen(t))) return -1; 
  return 0; }
static int write_transport3(const char *s, const char *t, const char *u)
{ if (MTA_SUCCESS!=mta->mta_write(mta,s,strlen(s))) return -1; 
  if (MTA_SUCCESS!=mta->mta_write(mta,t,strlen(t))) return -1; 
  if (MTA_SUCCESS!=mta->mta_write(mta,u,strlen(u))) return -1; 
  return 0; }
#define W1(x) write_transport(x)
#define W2(x,y) write_transport2(x,y)
#define W3(x,y,z) write_transport3(x,y,z)
#define C(x) do {int y=hdr_chr(&headers[x],'\n');headers[x].data[y]=0;} while (0)

static uostr_t *
encode_hdrstr(uostr_t *res,const char *str)
{
	const char *p;
	size_t l=strlen(str);
	char buf[24];
	int eightbit=0;
	char *q=buf;
	uostr_need(res,l); /* advise ... */
	uostr_cut(res,0);
	p=str;
	while (*p) {
		unsigned char c;
		if (*p=='\t') {eightbit=1;break;}
		c=(unsigned char)*p;
		if (c >127) {eightbit=1;break;}
		*q++=*p;
		if (q==buf+sizeof(buf)) { 
			if (!uostr_add_mem(res,buf,sizeof(buf))) return 0;
			q=buf;
		}
		p++;
	}
	if (!eightbit) { /* nothing problematic found */
		if (!uostr_add_mem(res,buf,q-buf)) return 0;
	} else {
		int wordlen=0;
		int did_head=0;
		uostr_cut(res,0); /* undo the string */
		/* btw: this code is never called in case of a message with a M-line */
		p=str;
		while (*p) {
			unsigned char c;
			if (!did_head) {
				if (!uostr_add_cstr(res,"=?iso-8859-1?Q?")) return 0;
				wordlen=15;
				did_head=1;
			}
			c=(unsigned char)*p;
			if (c=='\t' || c>127 || strchr("@.,;<>[]\\\"()?/=",c)) {
				char num[4];
				num[0]='=';
				if (c<10) {
					num[1]='0';
					str_ulong_base(num+2,c,16);
				} else
					str_ulong_base(num+1,c,16);
				if (!uostr_add_mem(res,num,3)) return 0;
				wordlen+=3;
			} else if (*p==' ') {
				if (!uostr_add_mem(res,"_",1)) return 0;
				wordlen++;
			} else {
				if (!uostr_add_mem(res,p,1)) return 0;
				wordlen++;
			}
			p++;
			if (wordlen+6>=72) {
				if (!uostr_add_mem(res,"?=",2)) return 0;
				if (*p && !uostr_add_mem(res," ",1)) return 0;
				wordlen=0; 
				did_head=0;
			}
		}
		if (did_head) {
			if (!uostr_add_mem(res,"?=",2)) return 0;
		}
	}
	if (!uostr_add_mem(res,"",1)) return 0;
	return res;
}

static uostr_t *
convert_address(uostr_t *s)
{
	static uostr_t t=UOSTR_INIT;
	static uostr_t u=UOSTR_INIT;
	char *p,*q;

	uostr_need(&t,s->len); /* hint, nothing more */
	uostr_cut(&t,0);

	p=s->data;
	q=s->data;
	while (1) {
		char *r,*end;
		char *w,*wstart;
		int needspace=0;
		while (*q!='\n') q++;
		if (!uostr_entumlauten_mem(&u,p,q-p)) return 0;
		if (!uostr_needmore(&t,u.len+1)) return 0;
		end=u.data+u.len;
		w=wstart=t.data+t.len;
		/* copy ascii address, deleting useless spaces */
		for (r=u.data;r<end;r++) {
			if (*r==' ') {
				needspace=1;
			} else {
				if (needspace && w>wstart && w[-1]!='@' && *r!='@') {	
					*w++='_';
					t.len++;
				}
				needspace=0;
				*w++=*r;
				t.len++;
			}
		}
		*w++='\n';
		t.len++;
		q++;
		p=q;
		if (q==s->data+s->len)
			break;
	}
	if (!uostr_add_mem(&t,"",1)) return 0;
	uostr_freedata(&u);
	u=*s;
	*s=t;
	t=u;
	uostr_freedata(&t);
	memset(&u,0,sizeof(u));
	return s;
}

static uostr_t *
output_addressheader(uostr_t *s,const char *tag)
{
	char *p,*q;
	size_t linelen=0;
	int first=1;
	if (!uostr_add_mem(s,"",1)) return 0; /* terminate it */
	W1(tag);
	p=q=s->data;
	while (1) {
		while (*q!='\n') q++;
		*q=0;
		if (linelen+q-p>75 && !first) {
			W1(",\n\t");
			linelen=1;
		} else if (!first) {
			W1(", ");
			linelen+=2;
		}
		W1(p);
		first=0;
		linelen+=q-p;
		q++;
		if (!*q) break;
	}
	W1("\n");
	uostr_cut(s,-1);
	return s;
}

static void
writeoutheader(void)
{
	mta_ret_t r;
	uostr_t s;s.data=0;
	if (!convert_address(&headers['V'])) goto oom;
	if (headers['S'].data && !convert_address(&headers['S'])) goto oom;
	if (headers['T'].data && !convert_address(&headers['T'])) goto oom;
	if (headers['G'].data) {
		uostr_t t;
		if (!uostr_entumlauten_mem(&s,headers['G'].data,headers['G'].len)) goto oom;
		t=s; s=headers['G']; headers['G']=t; /* swap them */
	}
	if (headers['F'].data) {
		uostr_t t;
		if (!uostr_entumlauten_mem(&s,headers['F'].data,headers['F'].len)) goto oom;
		t=s; s=headers['F']; headers['F']=t; /* swap them */
	}
	if (opt_dry_run) {
		r=mta_dummy.mta_open(&mta,0,0,0);
	} else if (messagetype==MT_NEWS) {
		r=mta_inews.mta_open(&mta,0,0,0);
	} else {
		if (!convert_address(&headers['A'])) goto oom;
		if (!uostr_add_mem(&headers['A'],"",1)) goto oom;
		if (!uostr_add_mem(&headers['V'],"",1)) goto oom;
		r=mta_qq.mta_open(&mta,0,headers['V'].data,headers['A'].data);
		uostr_cut(&headers['V'],-1);
		uostr_cut(&headers['A'],-1);
	}
	if (r!=MTA_SUCCESS) {
		mta=0;
		write_log("?kann Transportprogramm nicht starten");
		return;
	}
	C('#'); W3("Message-ID: <",headers['#'].data,">\n");
	if (headers['-'].data) {
		char *p,*end;
		int first=1;
		int pos=0;
		end=headers['-'].data+headers['-'].len;
		p=headers['-'].data;
		while (p!=end) {
			char *q=p;
			while (*q!='\n') q++;
			*q=0;
			if (q==end-1) {
				if (!first) W1("\n");
				W1("In-Reply-To: <");
				first=0;
			} else if (first) {
				W1("References: <");
				first=0;
			} else {
				W1(" <");
			}
			/* need not be exact here */
			if (pos && pos+headers['-'].len>70) {
				W1("\n\t"); 
				pos=0;
			}
			W2(headers['-'].data,">");
			pos+=headers['-'].len;

			p=q+1;
		}
		W1("\n");
	}
	W1("From: ");
	if (headers['N'].data) {
		C('N');
		if (!encode_hdrstr(&s,headers['N'].data)) goto oom;
		W3("\"",s.data,"\" ");
	}
	C('V');
	W3("<",headers['V'].data,">\n");

	if (messagetype==MT_NEWS) {
		char *p;
		W1("Newsgroups: ");
		for (p=headers['G'].data;p<headers['G'].data+headers['G'].len;) {
			char *q;
			q=memchr(p,'\n',headers['G'].len-(p-headers['G'].data));
			mta->mta_write(mta,p,q-p);
			p=q+1;
			if (p<headers['G'].data+headers['G'].len)
				mta->mta_write(mta,",",1);
		}
		mta->mta_write(mta,"\n",1);
	}
	C('W'); 
	if (!encode_hdrstr(&s,headers['W'].data)) goto oom;
	W3("Subject: ",s.data,"\n");
	{
		/* useless for news since inews forges the date header */
		size_t l;
		char timebuf[128];
		l=0;
		if (date.day<10) timebuf[l++]='0'; l+=str_ulong(timebuf+l,date.day); timebuf[l++]=' ';
		memcpy(timebuf+l,uo_monabbrev(date.mon),3); l+=3; timebuf[l++]=' ';
		l+=str_ulong(timebuf+l,date.year); timebuf[l++]=' ';
		if (date.hour<10) timebuf[l++]='0'; l+=str_ulong(timebuf+l,date.hour); timebuf[l++]=':';
		if (date.min<10) timebuf[l++]='0'; l+=str_ulong(timebuf+l,date.min); timebuf[l++]=':';
		if (date.sec<10) timebuf[l++]='0'; l+=str_ulong(timebuf+l,date.sec); timebuf[l++]=0;
		W3("Date: ",timebuf," +0000\n");
	}

	if (messagetype==MT_MAIL) {
		if (!output_addressheader(&headers['A'],"To: ")) goto oom;
	} else if (headers['D'].data) {
		C('D'); 
		switch(headers['D'].data[0]) {
		case 'M': W1("Distribution: mausnet\n"); break;
		case 'L': W1("Distribution: local\n"); break;
		}
	}
	if (headers['O'].data) { 
		C('O'); 
		if (!encode_hdrstr(&s,headers['O'].data)) goto oom;
		W3("Organization: ",s.data,"\n");
	}
	if (headers['S'].data) { 
		if (!output_addressheader(&headers['S'],"Sender: ")) goto oom;
	}
	if (headers['T'].data) { 
		if (!output_addressheader(&headers['T'],"Reply-To: ")) goto oom;
	}
	if (messagetype==MT_NEWS && headers['F'].data) {
		char *p;
		W1("Followup-To: ");
		for (p=headers['F'].data;p<headers['F'].data+headers['F'].len;) {
			char *q;
			q=memchr(p,'\n',headers['F'].len-(p-headers['F'].data));
			mta->mta_write(mta,p,q-p);
			p=q+1;
			if (p<headers['F'].data+headers['F'].len)
				mta->mta_write(mta,",",1);
		}
		mta->mta_write(mta,"\n",1);
	}
	if (headers['M'].data) { 
		int x=hdr_chr(&headers['M'],';');
		C('M');
		headers['M'].data[x]=0;
		if (headers['M'].data[x-1]==' ') headers['M'].data[x-1]=0;
		W3("Mime-Version: ",headers['M'].data,"\n");
		x++;
		while (x<(int) headers['M'].len 
			&& (headers['M'].data[x]==' '||headers['M'].data[x]=='\t'))
			x++;
		W2(headers['M'].data+x,"\n");
	} else
		W1("\n");
	uostr_freedata(&s);
	return;
  oom: mta->mta_abort(mta);
  	uostr_freedata(&s);
	messagetype=MT_JUNK;
}

static inline 
int count_lines(uostr_t *s)
{
	char *p;
	int count=0;
	if (!s->data) return 0;
	if (!s->len) return 0;
	for (p=s->data;p<s->data+s->len;p++)
		if (*p=='\n') 
			count++;
	return count;
}


/* go through an address list and qualify (+hostname if needed, +domain if needed) the addresses */
static int 
qualify(uostr_t *s)
{
	char *p;
	uostr_t n;
	n.data=0;
	for (p=s->data;p<s->data+s->len;) {
		char *q;
		/* copy localpart */
		/* go to @, skip spaces */
		q=p; while (*q!='\n' && *q!='@') { if (*p==' ' || *p=='\t') p++; q++; }
		if (q!=p && !uostr_add_mem(&n,p,q-p)) goto oom;
			/* delete useless space at end of localpart */
		while (n.len && (n.data[n.len-1]==' ' || n.data[n.len-1]=='\t')) n.len--;
		if (!uostr_add_char(&n,'@')) goto oom;
		if (*q=='@') q++;
		while (*q=='\t' || *q==' ') q++;
		/* check domain part */
		p=q;
		while (*q!='\n' && *q!='.') q++;
		if (*q=='\n') { /* a single host name, without domain, or no hostname at all */
			if (q-p && !uostr_add_mem (&n,p,q-p)) goto oom;
			else if (q==p) {
				if (!uostr_add_cstr (&n,UOCONF_DATA(&c_boxkuerzel))) goto oom;
			}
			if (!uostr_add_mem (&n,".",1)) goto oom;
			if (!uostr_add_cstr (&n,UOCONF_DATA(&c_defaultdomain))) goto oom;
			if (!uostr_add_mem (&n,"\n",1)) goto oom;
			p=q+1;
			continue;
		}
		/* something else */
		while (*q!='\n') q++;
		if (!uostr_add_mem(&n,p,q-p+1)) {uostr_freedata(&n);goto oom;}
		p=q+1;
	}
	if (!uostr_dup_uostr(s,&n)) goto oom;
	uostr_freedata(&n);
	return 1;
  oom:
    uostr_freedata(&n);
	errno=ENOMEM;
	return 0;
}

static int 
exact_one(char c)
{
	switch (count_lines(&headers[(int)c])) {
	case 0: return 0;
	case 1: return 1;
	default: return 0;
	}
}

static int
get_mtzeit(uostr_t *s, uo_sec70_t t)
{
	uo_datetime_t dt;
	char numbuf[STR_ULONG];
	uo_sec702dt(&dt,&t);
	str_ulong(numbuf,dt.year); if (!uostr_add_cstr(s,numbuf)) goto oom;
	dt.mon++;
	if (dt.mon<10 && !uostr_add_char(s,'0')) goto oom;
	str_ulong(numbuf,dt.mon); if (!uostr_add_cstr(s,numbuf)) goto oom;
	if (dt.day<10 && !uostr_add_char(s,'0')) goto oom;
	str_ulong(numbuf,dt.day); if (!uostr_add_cstr(s,numbuf)) goto oom;
	if (dt.hour<10 && !uostr_add_char(s,'0')) goto oom;
	str_ulong(numbuf,dt.hour); if (!uostr_add_cstr(s,numbuf)) goto oom;
	if (dt.min<10 && !uostr_add_char(s,'0')) goto oom;
	str_ulong(numbuf,dt.min); if (!uostr_add_cstr(s,numbuf)) goto oom;
	if (dt.sec<10 && !uostr_add_char(s,'0')) goto oom;
	str_ulong(numbuf,dt.sec); if (!uostr_add_cstr(s,numbuf)) goto oom;
    return 0;
 oom: errno=ENOMEM; return -1;
}

static void
dispatch_type(int got_colon)
{
	enum idtype { FQDN_ID,USER_ID,SPEC_ID } idtype;
	int num;
	messagetype=MT_JUNK;
	if (!exact_one('#')) {write_log("?nicht exakt eine #-Zeile"); return; }
	/* check id type */
	{
		int digits=0,ats=0,rest=0,uppers=0,lowers=0;
		char *p;
		for(p=headers['#'].data;p<headers['#'].data+headers['#'].len-1;p++) {
			if (*p>='0' && *p<='9') digits++;
			else if (*p=='@') ats++;
			else if (*p>='a' && *p<='z') lowers++;
			else if (*p>='A' && *p<='Z') uppers++;
			else rest++;
		}
		idtype=SPEC_ID;
		if (uppers && !lowers && !rest && !ats && !digits) idtype=SPEC_ID;
		else {
			if (ats) idtype=FQDN_ID;
			else idtype=USER_ID;
		}
		if (ats>1) { write_log("?zu viele @ in #-Zeilen"); return; }
	}
	/* wenn echte ID, B-Zeile, aber kein Doppelpunkt gekommen sind dann ist dies eine
	 * Statusmeldung
	 */
	if (headers['B'].data && idtype==FQDN_ID && !got_colon) {
		messagetype=MT_STATUS;
		goto got_typ;
	}
	if (idtype==SPEC_ID) {
		/* XXX todo */
		write_log("!ignoriert");
		return;
	}
	if (headers['B'].data) {
		write_log("?B-Zeile nicht erlaubt");
		return;
	}
	if (headers['G'].data) {
		messagetype=MT_NEWS;
		goto got_typ;
	}
	if (headers['A'].data) {
		messagetype=MT_MAIL;
		goto got_typ;
	}
	if (headers['-'].data) {
		write_log("?Siehe http://tirka.ohse.de/uwe/mausnet/probleme/gluecksmail.html");
		return;
	}
	write_log("?Wo sind A- oder G-Zeile?");
	return;
  got_typ:
	/* we double check some things here, because someday we might get a *-Line giving
	 * us the type of the message.
	 1G/is_u*/

	if (messagetype==MT_STATUS) {
		/* throw away, we can't do something about the brokenness */
		write_log("?Siehe http://tirka.ohse.de/uwe/mausnet/probleme/statusdesaster.html");
		messagetype=MT_JUNK;
		return;
	}
	/* demand real Reference IDs for news articles */
	if (messagetype==MT_NEWS && headers['-'].data) {
		char *p,*end;
		/* i could do somewhat more strict checking here, but this is not 
		 * nice: the user usually just sends back the garbage he got.
		 */
		for (end=headers['-'].data+headers['-'].len,p=headers['-'].data;p!=end;p++) {
			char *q;
			for (q=p;*q!='\n';q++) {
				if (*q=='@') break;
			}
			if (q==p||*q=='\n'||q[1]=='\n') { /* no @ found */
				write_log("?illegale '-'-Zeile");
				messagetype=MT_JUNK;
				return;
			}
			while (*q!='\n') q++;
			p=q;
		}
	}
	if (messagetype==MT_MAIL) {
		if (!headers['A'].data) {
			write_log("?A-Zeile vorgeschrieben"); 
			messagetype=MT_JUNK;
			return;
		} 
		/* qualify will do the rest */
	} else { /* MT_NEWS */
		if (headers['A'].data) {
			/* braindead fido: a useless "To"-Line. Don`t allow such stupidness ... */
			write_log("?A und G nicht erlaubt"); 
			messagetype=MT_JUNK;
			return;
		}
	}

	/* W */
	if (!exact_one('W')) { messagetype=MT_JUNK; write_log("?nicht exakt eine W-Zeile"); return; }
	/* O Organisation */
	num=count_lines(&headers['O']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele O-Zeilen");return;}
	/* D Distribution */
	num=count_lines(&headers['D']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele D-Zeilen");return;}
	if (num==1 && messagetype!=MT_NEWS) {
		messagetype=MT_JUNK;write_log("?D-Zeile hier nicht erlaubt");return;
	}
	if (num && (headers['D'].len>2 /* char plus \n  */
		|| !strchr("MNL",headers['D'].data[0]))) {
		messagetype=MT_JUNK;write_log("?illegale D-Zeile");return;
	}
	if (headers['D'].len==2 && headers['D'].data[0]=='M') {
		char *p;
		char *q;
		for (p=headers['G'].data,q=headers['G'].data+headers['G'].len;p && p<q-1;) {
			if (q-p>=4 && 0!=strncasecmp(p,"maus",4)) {
				write_log("?Distribution MausNet in nicht-maus-Gruppe");
				write_log("?Siehe http://tirka.ohse.de/uwe/mausnet/probleme/distribution.html");
				messagetype=MT_JUNK;
				return;
			}
			p=((char *)memchr(p,'\n',q-p))+1;
		}
	}
	/* Y */
	num=count_lines(&headers['Y']);
	if (num) { messagetype=MT_JUNK;write_log("?Y-Zeile nicht erlaubt");return;}
	/* X Flags */
	num=count_lines(&headers['X']);
	if (num) { messagetype=MT_JUNK;write_log("?X-Zeile nicht erlaubt");return;}
	/* I */
	num=count_lines(&headers['I']);
	if (num) { messagetype=MT_JUNK;write_log("?I-Zeile nicht erlaubt");return;}
	/* R */
	num=count_lines(&headers['R']);
	if (num) { messagetype=MT_JUNK;write_log("?R-Zeile nicht erlaubt");return;}

	/* C */
	num=count_lines(&headers['C']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele C-Zeilen");return;}
	if (num==1 && headers['M'].data) { write_log("!C-Zeile ignoriert, M-Zeile hat Vorrang");}

	/* M */
	num=count_lines(&headers['M']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele M-Zeilen");return;}
	if (num) {
		/* 1.0 ; content-type=... */
		int x=hdr_chr(&headers['M'],'\n');
		headers['M'].data[x]=0;
		if (0!=memcmp(headers['M'].data,"1.0 ;",5))
			{ messagetype=MT_JUNK;write_log("?illegale M-Zeile");return;}
		if (strstr(headers['M'].data,"really too many parameters"))
			/* 1. i don't want to deal with that.
			 * 2. i've still to see a single use of it (remember: it's a users
			 *    infile, use of MIME in mausnet is still low).
			 */
			{ messagetype=MT_JUNK;write_log("?bergewichtige M-Zeile");return;}
		headers['M'].data[x]='\n';
	}
	
	/* V */
	num=count_lines(&headers['V']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele V-Zeilen");return;}
	if (num==1 && headers['S'].data) { 
		messagetype=MT_JUNK;write_log("?V und S nicht erlaubt");return;}
	if (num==0) { 
		if (!UOCONF_DONE(&c_realname)) {
			messagetype=MT_JUNK;write_log("?Was ist Ihr Realname?");return;
		}
		if (!uostr_add_cstr(&headers['V'],UOCONF_DATA(&c_realname))) goto oom;
		if (!uostr_add_cstr(&headers['V'],"\n")) goto oom;
		/* qualify() will do the rest */
	} else { /* num == 1 */
		if (!UOCONF_DONE(&c_realname)) {
			messagetype=MT_JUNK;write_log("?Was ist Ihr Realname?");return;
		}
		if (!uostr_add_cstr(&headers['S'],UOCONF_DATA(&c_realname))) goto oom;
		/* qualify() will do the rest */
		if (!uostr_add_cstr(&headers['S'],"\n")) goto oom;
	}
	if (0==qualify(&headers['V'])) goto oom;
	/* A */
	if (messagetype==MT_MAIL) {
		num=count_lines(&headers['A']);
		if (num==0) { messagetype=MT_JUNK;write_log("?zu wenige A-Zeilen");return;}
		if (0==qualify(&headers['A'])) goto oom;
	}
	/* S Sender */
	if (headers['S'].data) {
		num=count_lines(&headers['S']);
		if (num>1) { messagetype=MT_JUNK;write_log("?zu viele S-Zeilen");return;}
		if (0==qualify(&headers['S'])) goto oom;
	}
	/* T Reply-To */
	if (headers['T'].data) {
		num=count_lines(&headers['T']);
		if (num>1) { messagetype=MT_JUNK;write_log("?zu viele T-Zeilen");return;}
		if (0==qualify(&headers['T'])) goto oom;
	}
	/* K Kopienempfnger */
	if (headers['K'].data) {
		if (0==qualify(&headers['K'])) goto oom;
	}
	/* N Realname */
	num=count_lines(&headers['N']);
	if (num>1) { messagetype=MT_JUNK;write_log("?zu viele N-Zeilen");return;}
	if (num==0) { 
		if (!UOCONF_DONE(&c_realname)) {
			messagetype=MT_JUNK;write_log("?Was ist Ihr Realname?");return;
		}
		if (!uostr_add_cstr(&headers['N'],UOCONF_DATA(&c_realname))) goto oom;
		if (!uostr_add_char(&headers['N'],'\n')) goto oom;
	}
	num=count_lines(&headers['G']);
	if (messagetype==MT_MAIL && num) {
		write_log("?G-Zeilen in Mail nicht erlaubt"); messagetype=MT_JUNK; return;
	}
	if (messagetype==MT_NEWS && !num) {
		write_log("?keine G-Zeilen gefunden"); messagetype=MT_JUNK; return;
	}
	if (num > 4) {
		num=count_lines(&headers['F']);
		if (num==0 || num>3) {
			write_log("?Zu viele G-Zeilen, F-Zeilen nicht in 1..3"); messagetype=MT_JUNK; return;
		}
	}
	/* E Eingabezeit */
	num=count_lines(&headers['E']);
	if (num>1) {
		write_log("?zu viele E-Zeilen"); messagetype=MT_JUNK; return;
	}
	if (num==0) {
		uo_sec70_t t;
		t=uo_now()+3600; /* CET */
		uo_sec702dt(&date,&t);
	} else { /* num==1 */
		/* the following will fail if germany ever moves more from 3600 minutes away from UTC.
		 * Oh well, at that time other things will be more interesting than fixing this
		 * problem :-)
		 * Of course it also fails in the missing and duplicate hour of the year, but i simply
		 * don't care: i don't get enough information in the E-Line.
		 */
		const char *data;
#define DATA(x) (data[x]-'0')
		int of;
		uo_sec70_t t;
		of=hdr_chr(&headers['E'],'\n');
		if (of!=12 && of!=14) {
			write_log("?illegale E-Zeile"); messagetype=MT_JUNK; return;
		}
		data=headers['E'].data;
		date.year=DATA(0)*1000+DATA(1)*100+DATA(2)*10+DATA(3);
		date.mon=(DATA(4)*10+DATA(5))-1;
		date.day=DATA(6)*10+DATA(7);
		date.hour=DATA(8)*10+DATA(9);
		date.min=DATA(10)*10+DATA(11);
		if (of==14) {
			date.sec=DATA(12)*10+DATA(13);
		} else
			date.sec=0;
#undef DATA
		t=uo_dt2sec70(&date);
		if (t+14*86400 < uo_now()) {
			write_log("?Nachricht steinalt"); messagetype=MT_JUNK; return;
		}
	}
#ifdef ITZ
	/* determine charset conversion */
	do_charset_conversion=0;
	if (!headers['M'].data) {
		const char *charset;
		if (headers['C'].data) {
			C('C');
			charset=headers['C'].data;
		} else 
			charset=UOCONF_DATA(&c_usercharset);
		/* XXX hardcoded charset preference */
		/* i only know what the senders wants ... but i don't know anything 
		 * about the receiver! So i use some form of lowest common den. */
		if (0!=strcasecmp(charset,"iso-8859-1")) {
			if (-1==itz_get_charset_by_kuerzel(charset)) {
				uostr_t meld;meld.data=0;
				if (!uostr_dup_cstr(&meld,"?Zeichensatz `")) goto oom;
				if (!uostr_dup_cstr(&meld,charset)) goto oom;
				if (!uostr_dup_cstr(&meld,"` unbekannt")) goto oom;
				if (!uostr_0(&meld)) goto oom;
				write_log(meld.data);
				uostr_freedata(&meld);
				messagetype=MT_JUNK;
				return;
			} else {
				int i;
				uostr_t uwan;
				uwan.data=0;
				itz_setup_conversion(charset,"iso-8859-1");
				/* this is rather efficient, we don`t need to free()/malloc() often */
				for (i=0;i<128;i++) {
					char *data;
					size_t len;
					size_t size;
					if (i=='#' || i=='-') continue;
					if (!headers[i].data) continue;
					if (!itz_convert_uostr(&uwan,&headers[i])) goto oom;
					data=uwan.data;
					len=uwan.len;
					size=uwan.size;
					uwan.data=headers[i].data;
					uwan.len=headers[i].len;
					uwan.size=headers[i].size;
					headers[i].data=data;
					headers[i].len=len;
					headers[i].size=size;
				}
				uostr_freedata(&uwan);
				do_charset_conversion=1;
			}
		}
	}
#endif
	if (idtype!=FQDN_ID) {
		/* generate a new ID */
		static int lfd=-1;
		char numbuf[STR_ULONG];
		uo_sec70_t t;
		if (lfd==-1) lfd=getpid();
		t=uo_now()+3600; /* CET */
		uostr_fcut(&headers['#']);
		if (-1==get_mtzeit(&headers['#'],t)) goto oom;
		if (!uostr_add_cstr(&headers['#'],".far.beyond.1990.x")) goto oom;
		str_ulong(numbuf,lfd++);
		if (!uostr_add_cstr(&headers['#'],numbuf)) goto oom;
		if (!uostr_add_mem(&headers['#'],"@",1)) goto oom;
		if (!uostr_add_cstr(&headers['#'],myhostname)) goto oom;
		if (!uostr_add_mem(&headers['#'],"\n",1)) goto oom;
	}
	writeoutheader();
	return;
  oom: /* get that out of control flow! */
	write_log("?out of memory"); 
	messagetype=MT_JUNK;
	return;
}

int 
main(int argc, char **argv)
{
	char buf[128];
	char *slash;
	memset(headers,0,sizeof(headers));
	argv0=strrchr(argv[0],'/'); if (!argv0) argv0=argv[0]; else argv0++;

	uoio_assign_r(&input,0,read,0);
	uoio_assign_w(&o,1,write,0);
	uoio_assign_w(&e,2,write,0);
	uoio_assign_w(&io_log,4,write,0);

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out,
		"usage: revo [options] [channelname]\n", myopts,
		"\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");

	if (opt_LOG) {
		int logfd=open(opt_LOG,O_WRONLY|O_CREAT|O_APPEND,0600);
		if (logfd==-1) {
			log (LOG_ERR, "cannot open %s: %s",opt_LOG,strerror(errno));
			exit(2);
		}
		uoio_assign_w(&io_log,logfd,write,0);
		uoio_write_cstr(&io_log,":#EXPORT\r\n");
	}
	if (!action) {
		if (!argv[1]) {
			log (LOG_ERR, "channel not given");
			exit(2);
		}
		action=argv[1];
		argv++;
	}
	if (argv[1]) {
		log (LOG_ERR, "too many arguments");
		exit (2);
	}
	/* configfile lesen */
	if (do_conf (cnffname, action,setvarscount ? setvars : NULL,setvarscount))
		exit (1);

	if (gethostname(buf,sizeof(buf)-1)==-1) bailout(errno,"gethostname failed",0,0,0);
	buf[sizeof(buf)-1]=0;
	myhostname=malloc(strlen(buf)+1);
	if (!myhostname) goto oom;
	memcpy(myhostname,buf,strlen(buf)+1);


	uogetopt(argv0,PACKAGE,VERSION,&argc,argv,uogetopt_out,
		"usage: revo\n",myopts,"\nreport bugs to <uwe-revo@bulkmail.ohse.de>");

	if (uoconf_environ(&c_realname,"MT_REALNAME")) bailout(0,"out of memory",0,0,0);
	if (uoconf_firstline(&c_realname,".revo/realname",1,'#'))
		bailout(errno,"cannot read ",".revo/realname",0,0);
	if (uoconf_cstr(&c_realname,opt_realname)) bailout(0,"out of memory",0,0,0);

#ifdef ITZ
	if (uoconf_environ(&c_itz,"MT_ITZ")) bailout(0,"out of memory",0,0,0);
	if (uoconf_firstline(&c_itz,".revo/ITZ",1,'#'))
		bailout(errno,"cannot read ",".revo/ITZ",0,0);
	if (uoconf_cstr(&c_itz,opt_itz)) bailout(0,"out of memory",0,0,0);
	/* i enforce existance of an ITZ */
	if (uoconf_cstr(&c_itz,"ITZ")) bailout(0,"out of memory",0,0,0);
	if (-1==itz_parse(UOCONF_DATA(&c_itz))) {
		bailout(errno,"cannot read ", UOCONF_DATA(&c_itz), ": ",0);
	}
#endif

	if (uoconf_environ(&c_usercharset,"MT_USERCHARSET")) bailout(0,"out of memory",0,0,0);
	if (uoconf_firstline(&c_usercharset,".revo/usercharset",1,'#'))
		bailout(errno,"cannot read ",".revo/usercharset",0,0);
	if (uoconf_cstr(&c_usercharset,o_usercharset)) bailout(0,"out of memory",0,0,0);
	if (uoconf_cstr(&c_usercharset,"iso-8859-1")) bailout(0,"out of memory",0,0,0);

	if (uoconf_environ(&c_defaultdomain,"MT_DEFAULTDOMAIN")) bailout(0,"out of memory",0,0,0);
	if (uoconf_firstline(&c_defaultdomain,".revo/defaultdomain",1,'#'))
		bailout(errno,"cannot read ",".revo/defaultdomain",0,0);
	if (uoconf_cstr(&c_defaultdomain,o_defaultdomain)) bailout(0,"out of memory",0,0,0);
	if (uoconf_cstr(&c_defaultdomain,"maus.de")) bailout(0,"out of memory",0,0,0);

	if (uoconf_environ(&c_boxkuerzel,"MT_BOXKUERZEL")) bailout(0,"out of memory",0,0,0);
	if (uoconf_firstline(&c_boxkuerzel,".revo/boxkuerzel",1,'#'))
		bailout(errno,"cannot read ",".revo/boxkuerzel",0,0);
	if (uoconf_cstr(&c_boxkuerzel,opts.boxkuerzel)) bailout(0,"out of memory",0,0,0);
	if (uoconf_cstr(&c_boxkuerzel,o_boxkuerzel)) bailout(0,"out of memory",0,0,0);
	if (!UOCONF_DONE(&c_boxkuerzel)) bailout(0,"$BOXKUERZEL not set",0,0,0);

	while (1) {
		ssize_t len;
		char *s;
		char c1;
		unsigned char c;
		len=uoio_getchar(&input,&c1);
		c=(unsigned char)c1;
		if (len==-1) {
			uoio_write_cstrmulti(&e,argv0,": cannot read stdin: ",
				strerror(errno),0);
			uoio_flush(&e);
			_exit(1);
		}
		if (len==0) break;
		if (c=='\n') continue; /* silently ignore empty lines */
		len=uoio_getdelim_zc(&input,&s,'\n');
		if (len==-1) {
			uoio_write_cstrmulti(&e,argv0,": cannot read stdin: ",
				strerror(errno),0);
			uoio_flush(&e);
			_exit(1);
		}
		if (len==0) break;
		if (c>=128) continue; /* silently ignored */
		if (c==':') { 
			if (in_header)
				dispatch_type(1);
			if (s[len-1]!='\n') {
				write_log("?letzte Zeile unvollstndig");
				if (mta) mta->mta_abort(mta);
				mta=0;
			} else if (len<2 || s[len-2]!='\r') {
				write_log("?Zeile ohne CR/LF");
				if (mta) mta->mta_abort(mta);
				mta=0;
			} else if (mta) {
				if (len==3 && *s=='-' && headers['M'].data) {
					/* extended Header for MIME messages.
					 * what a hack ... but the only way.
					 */
					uostr_freedata(&headers['M']);
				} else  {
					if (do_charset_conversion) {
						static uostr_t uwan=UOSTR_INIT;
						uostr_t fake;
						fake.size=len;
						fake.data=s;
						fake.len=len-2;
						if (!itz_convert_uostr(&uwan,&fake)) goto oom;
						mta->mta_write(mta,uwan.data,uwan.len);
					} else {
						mta->mta_write(mta,s,len-2);
					}
				}
				mta->mta_write(mta,"\n",1);
			}
			in_header=0; 
			continue;
		}
		if (c=='#') { /* starts a new message */
			int i;
			if (in_header) /* last message didn't contain a colon line */
				dispatch_type(0);
			in_header=0;
			if (mta) {
				if (MTA_SUCCESS!=mta->mta_close(mta)) {
					write_log2("?Transportfehler: ",mta_errtxt);
				} else {
					write_log2("=",headers['#'].data);
				}
			}
			for (i=0;i<128;i++) uostr_freedata(&headers[i]);
		}
		s[len-1]=0;len--;
		if (s[len-1]=='\r') { s[len-1]=0;len--; } /* get rid of \r */
		if (c=='#' && !len) break; /* End */
		if (c=='#') in_header=1;
		if (c=='#') {
			static uostr_t x;
			uostr_xdup_char(&x,'#');
			uostr_xadd_cstr(&x,s);
			uostr_x0(&x);
			write_log(x.data);
		}
		uostr_add_mem(&headers[c],s,len);
		uostr_add_char(&headers[c],'\n'); /* separator */
		uoio_write_char(&e,c);
		uoio_write_cstr(&e,s);
		uoio_write_mem(&e,"\n",1);
	}
	if (in_header) /* last message didn't contain a colon line */
		dispatch_type(0);
	if (-1==uoio_flush(&io_log)) {
		uoio_write_cstrmulti(&e,argv0,": cannot flush logfilechannel: ",
			strerror(errno),0);
		uoio_flush(&e);
		_exit(1);
	}
	_exit(0);
  oom:
	write(2,argv0,strlen(argv0));
	write(2,": out of memory\n",16);
	if (mta) mta->mta_abort(mta);
	_exit(1);
	return (1); /* keep compiler quiet */
}
