/* rfc.h: prototype for RFC specific functions */
/*
    Copyright (C) 1994 1995 1996 1997 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#ifndef RFC_H
#define RFC_H

#include "metamsg.h"

/*
 * Routinen zum Handling von Nachrichten in RFC News/Mail-Format
 */

	/* 
	 * wandelt time_t in RFC 1123 kompatiblen Zeitstring. 
	 * -> "Fri, 13 Mar 1994 18:19:20 GMT" 
	 */
size_t rfc_make_date (
	char *rfcdatebuf, /* Puffer mit mind. 40 Zeichen Laenge */
	size_t buflen,    /* Laenge des Puffers */
	uo_sec70_t  /* zu wandelnde Zeit */);

	/* 
	 * Headerzeile nach 7 bit wandeln
	 */
size_t rfc_head27bit (
	BOOLEAN mime,    /* MIME verwenden? Oder Umlaute in Langform wandeln? */
	uostr_t *to,        /* gewandelter String */
	const char *from /* Original */);

	/* 
	 * Kodiert eine ganze Headerzeile nach MIME.
	 */
size_t mime_code_headline (
	uostr_t *outputstring,  /* Gewandelter String, MIME - moeglicherweise gross */
	const unsigned char *inputstring /* der zu wandelnde Text */ );

	/* Kodiert den Inhalt einer Headerzeile nach MIME */
size_t mime_code_headstr (
	uostr_t *outputstring,  /* Gewandelter String, MIME - moeglicherweise gross */
	const unsigned char *inputstring /* der zu wandelnde Text */ );

	/* 
	 * Sucht eine Headerzeile in einer Nachricht. Die Funktion ist nicht 
	 * die schnellste, moeglicherweise lohnt es sich bei mehr als zwei
	 * oder drei zu suchenden Zeilen, die Nachricht ganz zu wandeln.
	 * returniert Zeiger auf Kopie des Inhalts.
	 */
char *rfc_find_header (
	const char *article, /* Puffer mit der Nachricht darin */
	const char *headername /* Name der Headerzeile ohne Doppelpunkt */ );

	/*
	 * Eine Nachricht aus einer Datei lesen. Die Datei wird ganz gelesen
	 * und der gesamte Inhalt als eine Nachricht betrachtet.
	 */
int rfc_read_message (
	const char *fname,  /* Der Name der zu lesenden Datei */
	uostr_t * buf /* in diesen Puffer, der automatisch vergroessert wird */);

	/*
	 * Eine Nachricht aus einer Datei lesen. Die Datei wird ganz gelesen
	 * und der gesamte Inhalt als eine Nachricht betrachtet.
	 */
int rfc_read_message_fd (
	int fd,     /* aus dieser Datei lesen */
	uostr_t * buf /* in diesen Puffer - der automatisch vergroessert wird */);

	/*
	 * Eine Anzahl Bytes als eine Nachricht aus einer Datei lesen.
	 */
int rfc_read_message_bytes (
	int fd,     /* aus dieser Datei lesen */
	size_t bytes, /* Laenge der zu lesenden Nachricht in Bytes */
	uostr_t * buf   /* Puffer fuer die gelesene Nachricht */);

	/* 
	 * Eine Nachricht im RFC-Format (Mail/News) ins metamsg-Format 
	 * konvertieren. Der Inhalt des Puffers wird veraendert, eine Kopie
	 * des Bodys in "meta" erzeugt. M.a.W., der Inhalt von buf ist 
	 * danach unbrauchbar und brachliegend.
	 */
int rfc_to_meta (
	uostr_t * buf, /* die zu konvertierende Nachricht */
	metamsg_t * meta,  /* die konvertierte Nachricht */
	int is_mail /* 0: news, 1: mail */ );

	/*
	 * Eine Nachricht im metamsg-Format im RFC-Format in eine Datei ausgeben.
	 */
void rfc_print_message (
	metamsg_t * meta, /* auszugebende Nachricht */
	FILE * f,         /* in diese Datei schreiben */
	const char *crlf, /* gewuenschtes Zeilenende */
	int dotescape     /* Punkte am Zeilenanfang verdoppeln? */ );

	/* Adresse in s in einen Listeneintrag verwandeln und anhaengen */
int rfc_adress_to_metalist (
	int meckern,       /* Fehlermeldung auf stderr ausgeben? */
	metalist_t ** ml,  /* An diese Liste anhaengen */
	char *s            /* die zu konvertierende Adresse */ ); 

	/* convert an id-line to a metalist. Return 
	 * TRUE if the last one was ok. Set troubles to 0 or count
	 * of problems 
	 */
BOOLEAN rfc_ids_to_metalist( 
	metalist_t **ml,    /* append to that list */
	const char *s		/* die zu konvertierende Zeile */,
	int *troubles);

#endif
